#include <SPI_MSTransfer.h>
#include "Stream.h"
#include <SPI.h>
#include <Wire.h>
#include <deque>
#include <queue>

#define SPI_TRANSACTION SPISettings(24000000, MSBFIRST, SPI_MODE0)
SPI_MSTransfer *_slave_pointer; 
std::deque<std::vector<uint16_t>> SPI_MSTransfer::teensy_handler_queue; // slave callback queue
std::deque<std::vector<uint16_t>> SPI_MSTransfer::teensy_stm_queue; // slave queue that the master cherry picks from
std::deque<std::vector<uint16_t>> SPI_MSTransfer::teensy_master_queue; // dormant state, possibly removed if unused.
_slave_handler_ptr SPI_MSTransfer::_slave_handler = nullptr;
_master_handler_ptr SPI_MSTransfer::_master_handler = nullptr;

void SPI_MSTransfer::onTransfer(_slave_handler_ptr handler){
  if ( _master_access ) _master_handler = handler;
  if ( _slave_access ) _slave_handler = handler;
}

SPI_MSTransfer::SPI_MSTransfer(const char *data, uint8_t cs, SPIClass *SPIWire) {
  chip_select = cs; spi_port = SPIWire; _master_access = 1;
  ::pinMode(cs,OUTPUT); // make sure CS is OUTPUT before deassertion.
  ::digitalWrite(cs,HIGH); // deassert the CS way before SPI initializes :)
  if ( !strcmp(data,"Serial") ) { _serial_port_identifier = 0; }
  else if ( !strcmp(data,"Serial1") ) { _serial_port_identifier = 1; }
  else if ( !strcmp(data,"Serial2") ) { _serial_port_identifier = 2; }
  else if ( !strcmp(data,"Serial3") ) { _serial_port_identifier = 3; }
  else if ( !strcmp(data,"Serial4") ) { _serial_port_identifier = 4; }
  else if ( !strcmp(data,"Serial5") ) { _serial_port_identifier = 5; }
  else if ( !strcmp(data,"Serial6") ) { _serial_port_identifier = 6; }






  else if ( !strcmp(data,"Wire") ) { wire_port = 0; }
  else if ( !strcmp(data,"Wire1") ) { wire_port = 1; }
  else if ( !strcmp(data,"Wire2") ) { wire_port = 2; }
  else if ( !strcmp(data,"Wire3") ) { wire_port = 3; }
  else if ( !strcmp(data,"SPI1") ) { remote_spi_port = 1; }
  else if ( !strcmp(data,"SPI2") ) { remote_spi_port = 2; }
  else if ( !strcmp(data,"EEPROM") ) { eeprom_support = 1; }
  else if ( !strcmp(data,"SERVO") ) { servo_support = 1; }
  else if ( !strcmp(data,"FASTLED" ) ) { fastled_support = 1; }
}



void SPI_MSTransfer::assert() {
  spi_port->beginTransaction(SPI_TRANSACTION); ::digitalWriteFast(chip_select,LOW);
}
void SPI_MSTransfer::deassert() { ::digitalWriteFast(chip_select,HIGH); spi_port->endTransaction(); }
bool SPI_MSTransfer::command_ack_response(uint16_t *data, uint32_t len) {
  uint8_t resend_count = 0; uint32_t timeout = millis();
  while ( spi_port->transfer16(0xFFA5) != 0xA506 ) {
    if ( millis() - timeout > 100 ) {
      if ( resend_count < 3 ) {
        resend_count++; deassert(); assert(); Serial.println("Resending packet, CRC fail / Out of Sync");
        for ( uint16_t i = 0; i < data[1]; i++ ) { delayMicroseconds(1); spi_port->transfer16(data[i]); }
        timeout = millis();
      }
      else {
        deassert(); return 0;
      }
    }
  }
  return 1;
}
void SPI_MSTransfer::digitalWriteFast(uint8_t pin, bool state) { digitalWrite(pin, state); }
void SPI_MSTransfer::digitalWrite(uint8_t pin, bool state) {
  if ( _slave_access ) return;
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++; // HEADER
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
  data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
  data[data_pos] = ((uint16_t)(pin << 8) | state); checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return; // RECEIPT ACK
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
      for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return;
      }
    }
  }
  deassert(); return;
}
bool SPI_MSTransfer::digitalReadFast(uint8_t pin) { return digitalRead(pin); }
bool SPI_MSTransfer::digitalRead(uint8_t pin) {
  if ( _slave_access ) return 0;
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++; // HEADER
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
  data[data_pos] = 0x0001; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
  data[data_pos] = pin; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return 0; // RECEIPT ACK
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
      for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return buf[2];
      }
    }
  }
  deassert(); return 0;
}
void SPI_MSTransfer::pinMode(uint8_t pin, uint8_t state) {
  if ( _slave_access ) return;
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++; // HEADER
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
  data[data_pos] = 0x0002; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
  data[data_pos] = ((uint16_t)(pin << 8) | state); checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return; // RECEIPT ACK
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
      for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return;
      }
    }
  }
  deassert(); return;
}
uint16_t SPI_MSTransfer::transfer16(uint16_t *buffer, uint16_t length, uint16_t packetID, bool fire_and_forget) {
  if ( _slave_access ) {
    uint16_t data[6+length], checksum = 0, data_pos = 0;
    data[data_pos] = 0xAA55; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = length; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = packetID; checksum ^= data[data_pos]; data_pos++;
    for ( uint16_t i = 0; i < length; i++ ) { data[data_pos] = buffer[i]; checksum ^= data[data_pos]; data_pos++; }
    data[data_pos] = checksum;
    std::vector<uint16_t> _vector(data, data + sizeof(data)/2); teensy_stm_queue.push_back(_vector);
    if ( teensy_stm_queue.size() > 32 ) teensy_stm_queue.pop_front();
    return packetID;
  }
  if ( _master_access ) {
    assert();
    uint16_t data[6+length], checksum = 0, data_pos = 0;
    data[data_pos] = ( fire_and_forget ) ? 0x9244 : 0x9243; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = length; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = packetID; checksum ^= data[data_pos]; data_pos++;
    for ( uint16_t i = 0; i < length; i++ ) { data[data_pos] = buffer[i]; checksum ^= data[data_pos]; data_pos++; }
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( fire_and_forget ) { deassert(); return data[4]; } // F&F, RETURN PACKETID
    if ( !command_ack_response(data, sizeof(data) / 2) ) return 0; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return buf[2];
        }
      }
    }
    deassert();
  }
  return 0;
}
void SPI_MSTransfer::begin(uint32_t baudrate) { // set remote serial port and baud
  if ( _slave_access ) return;
  if ( _serial_port_identifier != -1 ) {
    assert();
    uint16_t data[7], checksum = 0, data_pos = 0;
    data[data_pos] = 0x3235; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = _serial_port_identifier; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = baudrate>>16; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = baudrate; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return;
        }
      }
    }
    deassert(); return;
  }
}
int SPI_MSTransfer::read(void) {
  if ( _slave_access ) return -1;
  if ( _serial_port_identifier != -1 ) {
    assert();
    uint16_t data[5], checksum = 0, data_pos = 0;
    data[data_pos] = 0x3235; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0001; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = _serial_port_identifier; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return -1; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return (int16_t)buf[2];
        }
      }
    }
    deassert(); return -1;
  }
}
int SPI_MSTransfer::peek() {
  if ( _slave_access ) return -1;
  if ( _serial_port_identifier != -1 ) {
    assert();
    uint16_t data[5], checksum = 0, data_pos = 0;
    data[data_pos] = 0x3235; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0003; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = _serial_port_identifier; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return -1; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return (int16_t)buf[2];
        }
      }
    }
    deassert(); return -1;
  }
}
int SPI_MSTransfer::available(void) {
  if ( _slave_access ) return 0;
  if ( _serial_port_identifier != -1 ) {
    assert();
    uint16_t data[5], checksum = 0, data_pos = 0;
    data[data_pos] = 0x3235; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0002; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = _serial_port_identifier; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return 0; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return buf[2];
        }
      }
    }
    deassert(); return 0;
  }
}
size_t SPI_MSTransfer::write(const uint8_t *buf, size_t size) {
  if ( _slave_access ) return 0;
  if ( _serial_port_identifier != -1 ) {
    assert();
    uint16_t data[6+size], checksum = 0, data_pos = 0;
    data[data_pos] = 0x3235; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0004; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = _serial_port_identifier; checksum ^= data[data_pos]; data_pos++;
    data[data_pos] = size; checksum ^= data[data_pos]; data_pos++;
    for ( uint16_t i = 0; i < size; i++ ) { data[data_pos] = buf[i]; checksum ^= data[data_pos]; data_pos++; }
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return 0; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          deassert(); return buf[2];
        }
      }
    }
    deassert(); return 0;
  }
}


































































































void SPI_MSTransfer::begin() { // set remote I2C/SPI ports
}
uint16_t SPI_MSTransfer::analogRead(uint8_t pin) {
}
void SPI_MSTransfer::analogWrite(uint8_t pin, uint16_t value) {
}
void SPI_MSTransfer::analogReadResolution(uint8_t value) {
}
void SPI_MSTransfer::analogWriteResolution(uint8_t value) {
}
void SPI_MSTransfer::beginTransmission(uint8_t addr) {
}
void SPI_MSTransfer::endTransmission() {
}
void SPI_MSTransfer::requestFrom(uint8_t address, uint8_t bytes) {
}
uint8_t SPI_MSTransfer::transfer(uint8_t data) {
}
uint16_t SPI_MSTransfer::transfer16(uint16_t data) {
}
bool SPI_MSTransfer::online() {
}
void SPI_MSTransfer::software_reset() {
}
uint8_t SPI_MSTransfer::status_update() {
}
void SPI_MSTransfer::beginTransaction(uint32_t baudrate, uint8_t msblsb, uint8_t dataMode) { // set remote serial port and baud
}
void SPI_MSTransfer::endTransaction() { // end remote SPI transaction
}
int SPI_MSTransfer::read(int addr) {
}
uint16_t SPI_MSTransfer::length() {
}
size_t SPI_MSTransfer::write(int addr, uint8_t data) {
}
void SPI_MSTransfer::update(int addr, uint8_t data) {
}
uint32_t SPI_MSTransfer::crc() {
}
//void SPI_MSTransfer::show(uint8_t pin, CRGB *array, uint16_t array_length) { // update Fastled pixels
//}
size_t SPI_MSTransfer::print(const char *p) {
  while (*p) {
    char c = *p++;
    write(c);
  }
  return 0;
}
size_t SPI_MSTransfer::println(const char *p) {
  while (*p) {
    char c = *p++;
    write(c);
  }
  write('\n');
  return 0;
}
void SPI_MSTransfer::flush() {
}














































































SPI_MSTransfer::SPI_MSTransfer(const char *data, const char *mode) {
  if ( !strcmp(data,"SLAVE") && !strcmp(mode,"STANDALONE") ) {
    ::pinMode(13, OUTPUT);  // detach pin13 from SPI0 for led use.
    SIM_SCGC6 |= SIM_SCGC6_SPI0; // enable slave clock
    SPI0_MCR |= SPI_MCR_HALT | SPI_MCR_MDIS; // stop
    SPI0_CTAR0_SLAVE = SPI_CTAR_FMSZ(15) & SPI0_CTAR0_SLAVE & (~(SPI_CTAR_CPOL | SPI_CTAR_CPHA) | 0x00 << 25);
    SPI0_RSER = 0x00020000;
    CORE_PIN14_CONFIG = PORT_PCR_MUX(2);
    CORE_PIN11_CONFIG = PORT_PCR_DSE | PORT_PCR_MUX(2);
    CORE_PIN12_CONFIG = PORT_PCR_MUX(2);
    CORE_PIN2_CONFIG =  PORT_PCR_PS | PORT_PCR_MUX(2); // this uses pin 2 for the CS so Serial2 can be used instead.
    SPI0_MCR &= ~SPI_MCR_HALT & ~SPI_MCR_MDIS; // start
    NVIC_ENABLE_IRQ(IRQ_SPI0); // enable CS IRQ
    _slave_pointer = this; _slave_access = 1;
  }
} 
void spi0_isr(void) {
  static uint16_t spi_data[DATA_BUFFER_MAX]; uint16_t buffer_pos = 0;
  while ( !(GPIOD_PDIR & 0x01) ) {
    if ( SPI0_SR & 0xF0 ) {
      spi_data[buffer_pos] = SPI0_POPR;
      if ( buffer_pos++ >= DATA_BUFFER_MAX ) break; // BUFFER LENGTH PROTECTION
    }
    if ( spi_data[1] && buffer_pos >= spi_data[1] ) {
      _slave_pointer->SPI_MSTransfer::process_slave_data(spi_data, spi_data[1]); break;
    }
  }
  SPI0_SR |= SPI_SR_RFDF;
}
void SPI_MSTransfer::process_slave_data(uint16_t *data, uint16_t len) {
  uint16_t process_crc = 0;
  if ( data[0] == 0x9244 ) {
    std::vector<uint16_t> _vector(data, data + data[1]); teensy_handler_queue.push_back(_vector); return;
  }
  for ( uint16_t i = 0; i < len - 1; i++ ) {
    if (GPIOD_PDIR & 0x01) return; process_crc ^= data[i];
  }
//    for ( uint16_t i = 0; i < len; i++ ) {
//      Serial.print(data[i]); Serial.print(" ");
//    } Serial.println();
  if ( data[len - 1] != process_crc ) return; // CRC FAILED, DON'T PROCESS!
  while ( !(GPIOD_PDIR & 0x01) ) { // wait here until MASTER confirms ACK receipt
    if ( SPI0_SR & 0xF0 ) {
      SPI0_PUSHR_SLAVE = 0xA506; if ( SPI0_POPR == 0xFFA5 ) break;
    }
  }
  switch ( data[0] ) {
    case 0x9766: {
        switch ( data[2] ) {
          case 0x0000: { // SLAVE DIGITALWRITE(FAST) CONTROL
              ::digitalWriteFast((uint8_t)(data[3] >> 8), (uint8_t)data[3]);
              uint16_t checksum = 0, buf_pos = 0, buf[] = { 0xAA55, 0, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0001: { // SLAVE DIGITALREAD(FAST) CONTROL
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, ::digitalReadFast((uint8_t)data[3]), 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0002: { // SLAVE PINMODE CONTROL
              ::pinMode((uint8_t)(data[3] >> 8), (uint8_t)data[3]);
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
        }
        break;
      }
    case 0x9243: { // MASTER SENDS PACKET TO SLAVE QUEUE WITH CRC ACKNOWLEDGEMENT
        switch ( data[2] ) {
          case 0x0000: {
              std::vector<uint16_t> _vector(data, data + len); teensy_handler_queue.push_back(_vector);
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, data[4], 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
        }
        break;
      }
    case 0x9712: {
        switch ( data[2] ) {
          case 0x0000: {
              if ( teensy_stm_queue.size() > 0 ) { // IF QUEUES EXIST, ONE WILL BE DEQUEUED TO MASTER
                uint16_t checksum = 0, buf_pos = -1, buf[teensy_stm_queue.front()[1]];
                for ( uint16_t i = 0; i < (teensy_stm_queue.front()[1]); i++ ) buf[i] = teensy_stm_queue.front()[i];
                teensy_stm_queue.pop_front();
                while ( !(GPIOD_PDIR & 0x01) ) {
                  if ( SPI0_SR & 0xF0 ) {
                    if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                  }
                }
                return;
              }
              else { // OTHERWISE, SEND MINIMAL PACKET FOR NO QUEUES
                uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, 0, 0 }; buf[1] = sizeof(buf) / 2;
                for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                  if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
                }
                buf[(sizeof(buf) / 2) - 1] = checksum;
                while ( !(GPIOD_PDIR & 0x01) ) {
                  if ( SPI0_SR & 0xF0 ) {
                    if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                  }
                }
                return;
              }
            }
        }
        break;
      }
    case 0x3235: { // SERIAL PORT BEGIN METHOD
        switch ( data[2] ) {
          case 0x0000: {
              switch ( data[3] ) {
                case 0x0000: {
                    Serial.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0001: {
                    Serial1.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0002: {
                    Serial2.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0003: {
                    Serial3.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0004: {
                    Serial4.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0005: {
                    Serial5.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
                case 0x0006: {
                    Serial6.begin((uint32_t)data[4] << 16 | data[5]); break;
                  }
              }
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0001: {
              int16_t _read = 0;
              switch ( data[3] ) {
                case 0x0000: {
                    _read = Serial.read(); break;
                  }
                case 0x0001: {
                    _read = Serial1.read(); break;
                  }
                case 0x0002: {
                    _read = Serial2.read(); break;
                  }
                case 0x0003: {
                    _read = Serial3.read(); break;
                  }
                case 0x0004: {
                    _read = Serial4.read(); break;
                  }
                case 0x0005: {
                    _read = Serial5.read(); break;
                  }
                case 0x0006: {
                    _read = Serial6.read(); break;
                  }
              }
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, _read, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0002: {
              int16_t _available = 0;
              switch ( data[3] ) {
                case 0x0000: {
                    _available = Serial.available(); break;
                  }
                case 0x0001: {
                    _available = Serial1.available(); break;
                  }
                case 0x0002: {
                    _available = Serial2.available(); break;
                  }
                case 0x0003: {
                    _available = Serial3.available(); break;
                  }
                case 0x0004: {
                    _available = Serial4.available(); break;
                  }
                case 0x0005: {
                    _available = Serial5.available(); break;
                  }
                case 0x0006: {
                    _available = Serial6.available(); break;
                  }
              }
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, _available, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0003: {
              int16_t _peek = 0;
              switch ( data[3] ) {
                case 0x0000: {
                    _peek = Serial.peek(); break;
                  }
                case 0x0001: {
                    _peek = Serial1.peek(); break;
                  }
                case 0x0002: {
                    _peek = Serial2.peek(); break;
                  }
                case 0x0003: {
                    _peek = Serial3.peek(); break;
                  }
                case 0x0004: {
                    _peek = Serial4.peek(); break;
                  }
                case 0x0005: {
                    _peek = Serial5.peek(); break;
                  }
                case 0x0006: {
                    _peek = Serial6.peek(); break;
                  }
              }
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, _peek, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }
          case 0x0004: {
              uint16_t _written = 0; uint8_t _buf[data[4]];
              for ( uint16_t i = 0; i < data[4]; i++ ) _buf[i] = data[i+5];
              switch ( data[3] ) {
                case 0x0000: {
                    _written = Serial.write(_buf, data[4]); break;
                  }
                case 0x0001: {
                    _written = Serial1.write(_buf, data[4]); break;
                  }
                case 0x0002: {
                    _written = Serial2.write(_buf, data[4]); break;
                  }
                case 0x0003: {
                    _written = Serial3.write(_buf, data[4]); break;
                  }
                case 0x0004: {
                    _written = Serial4.write(_buf, data[4]); break;
                  }
                case 0x0005: {
                    _written = Serial5.write(_buf, data[4]); break;
                  }
                case 0x0006: {
                    _written = Serial6.write(_buf, data[4]); break;
                  }
              }
              uint16_t checksum = 0, buf_pos = -1, buf[] = { 0xAA55, 0, _written, 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint16_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) {
                if (GPIOD_PDIR & 0x01) return; checksum ^= buf[i];
              }
              buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  if ( buf_pos >= buf[1] ) buf_pos = 0; SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++;
                }
              }
              return;
            }















        }
        break;
      }








  }
}uint16_t SPI_MSTransfer::events() {
  if ( _master_access ) {
    assert();
    uint16_t data[4], checksum = 0, data_pos = 0;
    data[data_pos] = 0x9712; checksum ^= data[data_pos]; data_pos++; // HEADER
    data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++; // DATA SIZE
    data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++; // SUB SWITCH STATEMENT
    data[data_pos] = checksum;
    for ( uint16_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
    if ( !command_ack_response(data, sizeof(data) / 2) ) return 0; // RECEIPT ACK
    for ( uint8_t i = 0; i < 100; i++ ) {
      if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
        uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2; checksum = buf[0]; checksum ^= buf[1];
        for ( uint16_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); if ( i < buf[1]-1 ) checksum ^= buf[i]; }
        if ( checksum == buf[buf[1] - 1] ) {
          if ( buf[1] > 4 ) {
            uint16_t _slave_array[buf[3]]; for ( uint16_t i = 0; i < buf[3]; i++ ) _slave_array[i] = buf[i+5];
            deassert(); _master_handler(_slave_array, buf[3], buf[4]);
          }   
          deassert(); return 0;
        }
      }
    }
    deassert(); return 0;
  }
  if ( _slave_access ) {
    if ( teensy_handler_queue.size() > 0 ) {
      uint16_t _buf[teensy_handler_queue.front()[3]], checksum = 0;
      for ( uint16_t i = 0; i < (teensy_handler_queue.front()[1]) - 1; i++ ) {
        checksum ^= teensy_handler_queue.front()[i]; if ( i > 4 ) _buf[i - 5] = teensy_handler_queue.front()[i];
      }
      if ( checksum == teensy_handler_queue.front()[teensy_handler_queue.front()[1] - 1] && _slave_handler != nullptr ) {
        _slave_handler(_buf, teensy_handler_queue.front()[3], teensy_handler_queue.front()[4]);
      }
      teensy_handler_queue.pop_front(); return 0;
    }
  }
  return 0;
}











namespace std {
void __attribute__((weak)) __throw_bad_alloc() {
  Serial.println("Unable to allocate memory");
}
void __attribute__((weak)) __throw_length_error( char const*e ) {
  Serial.print("Length Error :"); Serial.println(e);
}
}
